@@ -14,18 +14,17 @@
 
 """Serialization/Deserialization for privsep.
 
-The wire format is a message length encoded as a simple unsigned int
-in native byte order (@I in struct.pack-speak), followed by that many
-bytes of UTF-8 JSON data.
-
+The wire format is a stream of msgpack objects encoding primitive
+python datatypes.  Msgpack 'raw' is assumed to be a valid utf8 string
+(msgpack 2.0 'bin' type is used for bytes).  Python lists are
+converted to tuples during serialization/deserialization.
 """
 
-import json
 import logging
 import socket
-import struct
 import threading
 
+import msgpack
 import six
 
 from oslo_privsep._i18n import _
@@ -33,9 +32,6 @@ from oslo_privsep._i18n import _
 
 LOG = logging.getLogger(__name__)
 
-_HDRFMT = '@I'
-_HDRFMT_LEN = struct.calcsize(_HDRFMT)
-
 
 try:
     import greenlet
@@ -55,14 +51,9 @@ class Serializer(object):
         self.writesock = writesock
 
     def send(self, msg):
-        buf = json.dumps(msg, ensure_ascii=False).encode('utf-8')
-
-        # json (the library) doesn't support push parsing and JSON
-        # (the format) doesn't include length information, so we can't
-        # decode without reading the entire input and blocking.  Avoid
-        # that by explicitly communicating the JSON message length
-        # first.
-        self.writesock.sendall(struct.pack(_HDRFMT, len(buf)) + buf)
+        buf = msgpack.packb(msg, use_bin_type=True,
+                            unicode_errors='surrogateescape')
+        self.writesock.sendall(buf)
 
     def close(self):
         # Hilarious. `socket._socketobject.close()` doesn't actually
@@ -74,28 +65,38 @@ class Serializer(object):
 class Deserializer(six.Iterator):
     def __init__(self, readsock):
         self.readsock = readsock
+        # (hberaud): Sometime msgpack's versions are not properly aligned with
+        # constraints and we need to stay compat with the available version.
+        # By example 1.0.0 dropped the encoding param but previous
+        # version doesn't define the raw params too.
+        # So here we use the right signature which fit our needs.
+        try:
+            self.unpacker = msgpack.Unpacker(
+                use_list=False,
+                encoding='utf-8',
+                unicode_errors='surrogateescape')
+        except TypeError:
+            self.unpacker = msgpack.Unpacker(
+                use_list=False,
+                raw=False,
+                strict_map_key=False,
+                unicode_errors='surrogateescape')
 
     def __iter__(self):
         return self
 
-    def _read_n(self, n):
-        """Read exactly N bytes.  Raises EOFError on premature EOF"""
-        data = []
-        while n > 0:
-            tmp = self.readsock.recv(n)
-            if not tmp:
-                raise EOFError(_('Premature EOF during deserialization'))
-            data.append(tmp)
-            n -= len(tmp)
-        return b''.join(data)
-
     def __next__(self):
-        try:
-            buflen, = struct.unpack(_HDRFMT, self._read_n(_HDRFMT_LEN))
-        except EOFError:
-            raise StopIteration
-
-        return json.loads(self._read_n(buflen).decode('utf-8'))
+        while True:
+            try:
+                return next(self.unpacker)
+            except StopIteration:
+                try:
+                    buf = self.readsock.recv(4096)
+                    if not buf:
+                        raise
+                    self.unpacker.feed(buf)
+                except socket.timeout:
+                    pass
 
 
 class Future(object):
@@ -126,6 +127,7 @@ class Future(object):
 
 class ClientChannel(object):
     def __init__(self, sock):
+        self.running = False
         self.writer = Serializer(sock)
         self.lock = threading.Lock()
         self.reader_thread = threading.Thread(
@@ -140,11 +142,18 @@ class ClientChannel(object):
 
     def _reader_main(self, reader):
         """This thread owns and demuxes the read channel"""
+        with self.lock:
+            self.running = True
         for msg in reader:
             msgid, data = msg
-            with self.lock:
-                assert msgid in self.outstanding_msgs
-                self.outstanding_msgs[msgid].set_result(data)
+            if msgid is None:
+                self.out_of_band(data)
+            else:
+                with self.lock:
+                    if msgid not in self.outstanding_msgs:
+                        raise AssertionError("msgid should in "
+                                             "outstanding_msgs.")
+                    self.outstanding_msgs[msgid].set_result(data)
 
         # EOF.  Perhaps the privileged process exited?
         # Send an IOError to any oustanding waiting readers.  Assuming
@@ -156,13 +165,19 @@ class ClientChannel(object):
         with self.lock:
             for mbox in self.outstanding_msgs.values():
                 mbox.set_exception(exc)
+            self.running = False
+
+    def out_of_band(self, msg):
+        """Received OOB message. Subclasses might want to override this."""
+        pass
 
     def send_recv(self, msg):
         myid = _get_thread_ident()
         future = Future(self.lock)
 
         with self.lock:
-            assert myid not in self.outstanding_msgs
+            if myid in self.outstanding_msgs:
+                raise AssertionError("myid shoudn't be in outstanding_msgs.")
             self.outstanding_msgs[myid] = future
             try:
                 self.writer.send((myid, msg))
